# Copyright 2021-2024 Nokia

import contextlib
import copy
import json
from enum import Enum, auto
from typing import List, Optional, Union

from .errors import *
from .errors import InvalidPathError, SrosMgmtError, make_exception
from .identifier import Identifier
from .model import Model, AModel
from .wrappers import Container, Leaf, LeafList
from .yang_type import *


class Token:
    def __init__(self):
        self.kind: Optional["ModelWalker._TokenKind"] = None
        self.value = ''


class ModelWalker:
    """API representation for walking model tree.

    .. Reviewed by TechComms 20210712
    """
    _expected_dds = (
        Model.StatementType.container_, Model.StatementType.list_,
        Model.StatementType.leaf_, Model.StatementType.leaf_list_,
        Model.StatementType.augment_, Model.StatementType.notification_,
        Model.StatementType.rpc_, Model.StatementType.input_,
        Model.StatementType.output_, Model.StatementType.choice_,
        Model.StatementType.case_, Model.StatementType.action_
    )
    _recursive_visited_dds = (
        Model.StatementType.module_, Model.StatementType.submodule_,
        Model.StatementType.uses_, Model.StatementType.grouping_,
        Model.StatementType.augment_
    )

    def __init__(self, model: Model, sros: bool):
        self.path: List[Model] = []
        self.keys = []
        while model.parent is not None:
            if model.data_def_stm in self._expected_dds or not self.path:
                self.path.insert(0, model)
                self.keys.insert(0, dict())
            model = model.parent
        self.model = model
        self._sros = sros

    @property
    def current(self):
        return self.path[-1] if len(self.path) > 0 else self.model

    @property
    def local_keys(self):
        return self.keys[-1]

    @property
    def is_root(self):
        return not self.path

    @property
    def is_state(self):
        return not self.current.config

    def go_to_parent(self):
        if not self.path:
            raise make_exception(pysros_err_cannot_call_go_to_parent)

        self.path.pop()
        self.keys.pop()

    def go_to_child(self, child_name: Union[str, Identifier]):
        child = self._get_child(child_name)
        self.path.append(child)
        self.keys.append(dict())
        return child

    def go_to(self, path: ModelPath, module: Optional[str] = None):
        for p in path._path:
            if not p.is_lazy_bound():
                try:
                    self.go_to_child(p)
                except SrosMgmtError as e:
                    raise InvalidPathError(*e.args) from None
                continue

            if p.name == '..':
                self.go_to_parent()
                continue

            #prefix = self.path[-1].name.prefix if self.path else None
            prefix = None

            try:
                if prefix:
                    self.go_to_child(Identifier(prefix, p.name))
                else:
                    self.go_to_child(p.name)
            except SrosMgmtError as e:
                raise InvalidPathError(*e.args) from None

    def check_child_field_value(self, name: Union[str, Identifier], value, *, json=False, strict=False):
        with self.visit_child(name):
            return self.check_field_value(value, json=json, strict=strict)

    def check_field_value(self, value, *, json=False, strict=False, is_convert=False, metadata=None):
        if self.get_dds() == Model.StatementType.leaf_list_:
            if not isinstance(value, list):
                raise make_exception(pysros_err_leaflist_should_be_list, type_name=value.__class__.__name__)
            return all(self.get_type().check_field_value(i, json=json, strict=strict, is_convert=is_convert, metadata=metadata) for i in value)
        elif self.get_dds() == Model.StatementType.leaf_:
            return self.get_type().check_field_value(value, json=json, strict=strict, is_convert=is_convert, metadata=metadata)
        else:
            assert False, "Checking field value for non-field walker"

    def get_parent(self):
        res = self.__class__(self.model, self._sros)
        if self.path:
            res.path = self.path[:-1]
            res.keys = self.keys[:-1]
        return res

    def get_child(self, child_name: Union[str, Identifier]):
        res = self.copy()
        res.path.append(self._get_child(child_name))
        res.keys.append(dict())
        return res

    def copy(self):
        res = self.__class__(self.model, self._sros)
        res.path = self.path[:]
        res.keys = copy.deepcopy(self.keys)
        return res

    def go_to_last_with_presence(self):
        while self.path and not self.has_explicit_presence():
            self.go_to_parent()

    def has_explicit_presence(self):
        types_with_presence = (
            Model.StatementType.leaf_, Model.StatementType.leaf_list_
        )
        return (
            self.get_dds() in types_with_presence
            or self.get_dds() == Model.StatementType.list_ and not self.has_missing_keys()
            or self.get_dds() == Model.StatementType.container_ and self.current.presence_container)

    def has_missing_keys(self):
        return (
            self.get_dds() == Model.StatementType.list_ and
            len(self.keys[-1]) != len(self.current.local_keys)
        )

    def dict_keys(self, value):
        for i in value.keys():
            if not isinstance(i, tuple):
                i = (i,)
            if (len(self.get_local_key_names()) != len(i)):
                return False
            if not all(self.check_child_field_value(key_name, key_val, strict=True) for key_name, key_val in zip(self.get_local_key_names(), i)):
                return False
        return all(isinstance(v, (dict, Container)) for v in value.values())

    def entry_keys(self, value):
        def correct_key(entry, key):
            if key in entry and self.check_child_field_value(key, entry[key], strict=True):
                return True
            return False

        local_keys = [Identifier(self.get_name().prefix, k) for k in self.get_local_key_names()]
        return all(correct_key(value, k) for k in local_keys)

    def as_model_type(self, s):
        assert self.get_dds() in (Model.StatementType.leaf_, Model.StatementType.leaf_list_)
        if isinstance(s, list):
            res = [self.get_type().to_value(p) for p in s]
        else:
            res = [self.get_type().to_value(s)]

        if self.get_dds() == Model.StatementType.leaf_:
            res = res[0]
        return res

    def dump_json(self):
        def _dump_json(self, parent_data):
            if self.get_dds() in self._expected_dds:
                data = {"name": self.get_name().name,
                        "dds":  self.get_dds().name}
                if self.get_dds() != Model.StatementType.leaf_:
                    data["children"] = []
                if self.get_dds() == Model.StatementType.leaf_:
                    data["type"] = self.get_type().json_name()
                parent_data["children"].append(data)
                parent_data = data

            children = list(self.current.children)
            while children:
                child = children.pop()
                if child.data_def_stm in (Model.StatementType.module_, Model.StatementType.submodule_):
                    children.extend(child.children)
                    continue
                if child.data_def_stm in (Model.StatementType.container_, Model.StatementType.list_, Model.StatementType.leaf_, Model.StatementType.leaf_list_, Model.StatementType.choice_, Model.StatementType.case_, ):
                    with self.visit_child(child.name.name):
                        _dump_json(self, parent_data)
        root_data = {"children": []}
        _dump_json(self, root_data)
        return json.dumps(root_data, indent=4)

    def as_child_model_type(self, child, s):
        with self.visit_child(child):
            return self.as_model_type(s)

    @contextlib.contextmanager
    def visit_child(self, child: Union[str, Identifier]):
        self.go_to_child(child)
        try:
            yield
        finally:
            self.go_to_parent()

    @contextlib.contextmanager
    def visit_parent(self):
        if not self.path:
            raise make_exception(pysros_err_cannot_call_go_to_parent)
        keys, model = self.keys.pop(), self.path.pop()
        try:
            yield
        finally:
            assert model.parent == self.current
            self.keys.append(keys)
            self.path.append(model)

    def get_child_type(self, child_name: Union[str, Identifier]):
        return self._get_child(child_name).yang_type

    def get_child_dds(self, child_name: Union[str, Identifier]):
        return self._get_child(child_name).data_def_stm

    def get_child_name(self, child_name: Union[str, Identifier]):
        return self._get_child(child_name).name

    def get_type(self):
        return self.current.yang_type

    def get_dds(self):
        return self.current.data_def_stm

    def get_name(self):
        return self.current.name

    def get_local_key_names(self):
        return self.current.local_keys

    def has_local_keys_specified(self):
        return len(self.get_local_key_names()) == len(self.local_keys)

    def has_local_key_named(self, name):
        return name in self.get_local_key_names()

    def has_field_named(self, name):
        return (
            self.has_child(name) and
            self.get_child_dds(name) in (
                Model.StatementType.leaf_, Model.StatementType.leaf_list_
            )
        )

    @property
    def is_leaflist(self):
        return self.get_dds() == Model.StatementType.leaf_list_

    @property
    def is_local_key(self):
        return self.get_name().name in self.get_parent().get_local_key_names()

    def has_child(self, child_name: Union[str, Identifier] = None):
        if child_name:
            try:
                self._get_child(child_name)
            except:
                return False
            else:
                return True
        else:
            return self.current.has_children

    def validate_get_filter(self, filter: Union[dict, Container]):
        if filter == Container({}):
            filter = {}
        if self.get_dds() in (Model.StatementType.list_, Model.StatementType.container_):
            if isinstance(filter, Container):
                filter = filter.data
            if not isinstance(filter, dict):
                raise make_exception(pysros_err_filter_should_be_dict)
            unwrap = lambda v: v.data if isinstance(v, Leaf) else v
            if any(unwrap(v) == '' for v in filter.values()):
                raise make_exception(pysros_err_filter_empty_string)
            for k, v in filter.items():
                with self.visit_child(k):
                    self.validate_get_filter(v)
        elif self.get_dds() in (Model.StatementType.leaf_, Model.StatementType.leaf_list_):
            if isinstance(filter, Leaf) and self.get_dds() == Model.StatementType.leaf_:
                filter = filter.data
            if isinstance(filter, LeafList) and self.get_dds() == Model.StatementType.leaf_list_:
                filter = filter.data
            if isinstance(filter, dict):
                if filter:
                    raise make_exception(pysros_err_filter_wrong_leaf_value, leaf_name=self.get_name().name)
            elif isinstance(filter, str):
                pass
            elif not self.check_field_value(filter, strict=True):
                raise make_exception(
                    pysros_err_incorrect_leaf_value,
                    leaf_name=self.get_name().name
                )

    class _TokenKind(Enum):
        string = auto()
        symbol = auto()

    @classmethod
    def _tokenize(cls, string):
        for i in cls._tokenize_(string):
            yield i

    @classmethod
    def _tokenize_(cls, string):
        res = []
        iterator = iter(string)
        while True:
            try:
                i = next(iterator)
            except StopIteration:
                if res:
                    yield (cls._TokenKind.string, "".join(res))
                return
            if i in "[]=/":
                if res:
                    yield (cls._TokenKind.string, "".join(res))
                    res = []
                yield (cls._TokenKind.symbol, i)
                continue
            if res:
                res.append(i)
                continue
            if i == '"':
                while True:
                    try:
                        i = next(iterator)
                        if i == '"':
                            break
                        res.append(i)
                    except StopIteration:
                        raise make_exception(pysros_err_unended_quoted_string)
                yield (cls._TokenKind.string, "".join(res))
                res = []
                continue

            if i == "'":
                while True:
                    try:
                        i = next(iterator)
                        if i == "'":
                            break
                        res.append(i)
                    except StopIteration:
                        raise make_exception(pysros_err_unended_quoted_string)
                yield (cls._TokenKind.string, "".join(res))
                res = []
                continue
            res.append(i)

    @classmethod
    def user_path_parse(cls, model_root, path, sros, *, accept_root=False, verify_keys=True):
        res = cls.path_parse(model_root, path, sros, verify_keys)
        assert isinstance(res, ModelWalker)
        if not accept_root and res.is_root:
            raise make_exception(pysros_err_root_path)
        return res

    @classmethod
    def path_parse(cls, model_root, path_string, sros, verify_keys):
        assert isinstance(verify_keys, bool)
        if not isinstance(path_string, str):
            raise make_exception(pysros_err_path_should_be_string)

        if path_string == "/":
            return cls(model_root, sros)

        if not path_string.startswith('/'):
            if not path_string:
                raise make_exception(pysros_err_empty_path)
            raise make_exception(pysros_err_not_found_slash_before_name)
        if path_string.endswith('/'):
            raise make_exception(pysros_err_invalid_identifier)

        correct_char = lambda c: 32 <= ord(c) <= 127 or c in '\t\r\n'
        if any(not correct_char(c) for c in path_string):
            raise make_exception(pysros_err_invalid_parse_error)

        iterator = iter(cls._tokenize(path_string))

        def next_token(*accepts, err,  **kwarg):
            """Verify a next token from path and return tuple(kind, token) if no errors.
                Otherwise, raise make_exception(err, **kwarg).

:param accepts: Contains a list of the following:
    - one of '/', '[', ']' or '=' - Function accepts specified symbol.
    - 'string' - Function accepts a string, not a symbol.
    - None - Function accepts end of string.

            .. Reviewed by TechComms 20210713
            """
            n = next(iterator, None)

            if n is None:
                if None in accepts:
                    return None, None
            elif n[0] == cls._TokenKind.symbol:
                if n[1] in accepts:
                    return n
            elif 'string' in accepts:
                return n
            raise make_exception(err, **kwarg)

        res = cls(model_root, sros)
        missing_keys = set()
        elem = 'root'

        while True:
            _, i = next_token('/', '[', None, err=pysros_err_not_found_slash_before_name)
            if i == None:
                if (
                    missing_keys
                    and len(missing_keys) != len(res.current.local_keys)
                ):
                    raise make_exception(pysros_err_missing_keys, element=elem)
                break
            if i == "/":
                if missing_keys:
                    raise make_exception(pysros_err_missing_keys, element=elem)
                _, elem = next_token('string', None, err=pysros_err_invalid_identifier)
                try:
                    res.go_to_child(elem)
                    missing_keys.update(res.current.local_keys)
                except Exception:
                    try:
                        res.go_to_child(Identifier(res.current.name.prefix, elem))
                        missing_keys.update(res.current.local_keys)
                    except Exception:
                        raise make_exception(
                            pysros_err_unknown_element,
                            element=elem) from None
                continue
            if i == "[":
                _, key_name = next_token(
                    'string', err=pysros_err_invalid_identifier
                )
                if verify_keys:
                    next_token('=', err=pysros_err_expected_equal_operator)
                    _, value = next_token(
                        'string', err=pysros_err_invalid_identifier
                    )
                next_token(']', err=pysros_err_expected_end_bracket)
                if key_name not in missing_keys:
                    try:
                        res._get_child(key_name)
                    except Exception:
                        raise make_exception(
                            pysros_err_unknown_key, key_name=key_name
                        )
                    raise make_exception(
                        pysros_err_cannot_specify_non_key_leaf
                    )
                missing_keys.remove(key_name)
                if verify_keys:
                    res.local_keys[key_name] = value
        return res

    @classmethod
    def tokenize_path(cls, path: str, absolute_schema_id: bool):
        # absolute schema id has form: absolute-schema-nodeid = 1*("/" node-identifier)
        # leaf-ref form allows also predicates and a ".." inside paths
        if not path.startswith(('/', '../')):
            raise make_exception(pysros_err_invalid_yang_path, path=path)

        _end = None, ''
        t = Token()

        iterator = iter(cls._tokenize_(path))

        def next_token(t):
            t.kind, t.value = next(iterator, _end)
        next_token(t)
        if t.value == '/':
            # skip '/' in absolute path
            next_token(t)

        while True:
            if t.kind != cls._TokenKind.string:
                raise make_exception(pysros_err_invalid_yang_path, path=path)
            yield t.value
            next_token(t)

            while t.value == '[' and not absolute_schema_id:
                while t.value not in ('', ']'):
                    next_token(t)
                    if t.kind is None:
                        return
                next_token(t)

            if t.value == '/':
                next_token(t)
                continue

            if t.kind is None:
                return

            raise make_exception(pysros_err_invalid_yang_path, path=path)

    def __str__(self):
        return "/" + "/".join(str(name.name) + "".join(f"[{key}={value}]" for key, value in keys.items()) for name, keys in zip(self.path, self.keys))

    def has_unique_child(self, child_name: str):
        try:
            self._get_child(child_name)
        except SrosMgmtError:
            return False
        return True

    def _get_child(self, child_name: Union[str, Identifier]):
        children = list(self.current.children)
        res = None
        while children:
            child = children.pop()
            if child.name == child_name and child.data_def_stm in self._expected_dds and self._is_allowed(child):
                if res is None:
                    res = child
                else:
                    raise make_exception(pysros_err_ambiguous_model_node)
            if child.data_def_stm in self._recursive_visited_dds:
                children.extend(child.children)
        if res is not None:
            return res
        raise make_exception(
            pysros_err_unknown_child,
            child_name=child_name,
            path=self._get_path()
        )

    def recursive_walk(self, *, enter_fnc=None, leave_fnc=None):
        children = list(self.current.children)
        while children:
            child = children.pop()
            if child.data_def_stm in self._expected_dds and self._is_allowed(child):
                self.path.append(child)
                self.keys.append(dict())
                enter_fnc and enter_fnc(self)
                self.recursive_walk(enter_fnc=enter_fnc, leave_fnc=leave_fnc)
                leave_fnc and leave_fnc(self)
                self.path.pop()
                self.keys.pop()
            if child.data_def_stm in self._recursive_visited_dds:
                children.extend(child.children)

    def _get_path(self):
        return " ".join(str(model.name) for model in self.path)

    def _is_allowed(self, model):
        return True

    def is_region_blocked_in_child(self, child_name: Union[str, Identifier]):
        return self._get_child(child_name).is_region_blocked


class DataModelWalker(ModelWalker):
    _expected_dds = (
        Model.StatementType.container_, Model.StatementType.list_,
        Model.StatementType.leaf_, Model.StatementType.leaf_list_,
        Model.StatementType.action_, Model.StatementType.input_,
        Model.StatementType.output_
    )
    _recursive_visited_dds = (
        Model.StatementType.module_, Model.StatementType.submodule_,
        Model.StatementType.uses_, Model.StatementType.augment_,
        Model.StatementType.choice_, Model.StatementType.case_
    )


class FilteredDataModelWalker(DataModelWalker):
    _expected_dds = (
        Model.StatementType.container_, Model.StatementType.list_,
        Model.StatementType.leaf_, Model.StatementType.leaf_list_
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config_only = False
        self.return_blocked_regions = False

    def _is_allowed(self, model):
        if self.config_only and not model.config:
            return False
        if not self.return_blocked_regions and model.is_region_blocked:
            return False
        if self._sros:
            return any(i in model.name.prefix for i in ("nokia", "openconfig"))
        return True

    def _construct_path_without_key_values(self):
        path = ""
        parent_module_name = None
        for node in self.path:
            current_module_name = node.name.prefix
            current_node_name   = f"{node.name.name}"
            if node.data_def_stm == AModel.StatementType.list_:
                current_node_name += "".join((f"[{key}]" for key in node.local_keys))
            if parent_module_name != current_module_name:
                path += f"/{current_module_name}:{current_node_name}"
                parent_module_name = current_module_name
            else:
                path += f"/{current_node_name}"
        yield path

    def iterate_children(self, *, enter_fnc=None, action_io):
        children = list(self.current.children)
        while children:
            child = children.pop()
            if child.data_def_stm in self._expected_dds and self._is_allowed(child):
                self.path.append(child)
                self.keys.append(dict())
                if enter_fnc:
                    yield from enter_fnc(self, action_io)
                yield from self.iterate_children(enter_fnc=enter_fnc, action_io=action_io)
                self.path.pop()
                self.keys.pop()
            if child.data_def_stm in self._recursive_visited_dds:
                children.extend(child.children)

    def _export_paths(self, args, action_io):
        if self.current is None:
            return
        if not self.is_root:
            if self.current.name.name in self.current.parent.local_keys:
                return
            if self.current.status:
                if action_io in ("input", "output"):
                    if len(self.path) > 1:
                        if any((node.data_def_stm == Model.StatementType.action_ for node in self.path)):
                            yield from self._construct_path_without_key_values()
                elif action_io == "action_only":
                    if len(self.path) > 1:
                        if self.path[-1].data_def_stm == Model.StatementType.action_:
                            yield from self._construct_path_without_key_values()
                else:
                    yield from self._construct_path_without_key_values()


class ActionInputFilteredDataModelWalker(FilteredDataModelWalker):
    _expected_dds = (
        Model.StatementType.container_, Model.StatementType.list_,
        Model.StatementType.leaf_, Model.StatementType.leaf_list_,
        Model.StatementType.action_
    )
    _recursive_visited_dds = (
        Model.StatementType.module_, Model.StatementType.submodule_,
        Model.StatementType.uses_, Model.StatementType.augment_,
        Model.StatementType.choice_, Model.StatementType.case_,
        Model.StatementType.input_
    )


class ActionOutputFilteredDataModelWalker(FilteredDataModelWalker):
    _expected_dds = (
        Model.StatementType.container_, Model.StatementType.list_,
        Model.StatementType.leaf_, Model.StatementType.leaf_list_,
        Model.StatementType.action_
    )
    _recursive_visited_dds = (
        Model.StatementType.module_, Model.StatementType.submodule_,
        Model.StatementType.uses_, Model.StatementType.augment_,
        Model.StatementType.choice_, Model.StatementType.case_,
        Model.StatementType.output_
    )

class JsonInstanceModelWalkerWithActionInput(ActionInputFilteredDataModelWalker):
    def export_paths(self):
        if self.current.data_def_stm in (AModel.StatementType.leaf_, AModel.StatementType.leaf_list_):
            if any((node.data_def_stm == Model.StatementType.action_ for node in self.path)):
                yield from self._construct_path_without_key_values()
        yield from self.iterate_children(enter_fnc=self._export_paths, action_io="input")


class JsonInstanceModelWalkerWithActionOutput(ActionOutputFilteredDataModelWalker):
    def export_paths(self):
        if self.current.data_def_stm in (AModel.StatementType.leaf_, AModel.StatementType.leaf_list_):
            if any((node.data_def_stm == Model.StatementType.action_ for node in self.path)):
                yield from self._construct_path_without_key_values()
        yield from self.iterate_children(enter_fnc=self._export_paths, action_io="output")


class JsonInstanceModelWalkerActionOnly(ActionInputFilteredDataModelWalker):
    def export_paths(self):
        yield from self.iterate_children(enter_fnc=self._export_paths, action_io="action_only")

class JsonInstanceDataModelWalker(FilteredDataModelWalker):
    def export_paths(self):
        if self.current.data_def_stm in (AModel.StatementType.leaf_, AModel.StatementType.leaf_list_):
            if all((node.data_def_stm != Model.StatementType.action_ for node in self.path)):
                yield from self._construct_path_without_key_values()
        yield from self.iterate_children(enter_fnc=self._export_paths, action_io=None)



# Copyright 2021-2024 Nokia

import operator
import collections

from abc import abstractmethod
from io import StringIO
from functools import total_ordering

from .errors import *
from .errors import make_exception
from .singleton import _Empty
from .yang_type import DECIMAL_LEAF_TYPE, INTEGRAL_LEAF_TYPE, YangUnion

__all__ = (
    "Action", "Container", "Leaf", "LeafList", "Annotation", "Annotations"
)

__doc__ = """This module contains wrappers describing the YANG
structure and metadata obtained from SR OS.

.. Reviewed by PLM 20211201
.. Reviewed by TechComms 20211202
"""


class Schema:
    """YANG schema supporting information associated with elements in the data structure.

    .. note:: :py:class:`pysros.wrappers.Schema` metadata is read-only.

    .. property:: module

       YANG module name from which this node originates.

       :rtype: str

    .. property:: namespace

       YANG module namespace from which this node originates.  This is in URN or URL format.

       :rtype: str

    .. property:: yang_type

       YANG data type.  This type is the derived base YANG type, for example, if a YANG node uses
       a ``typedef``, and that ``typedef`` is a ``uint8``, yang_type returns ``uint8``.

       :rtype: str, :py:class:`SchemaType`

    .. property:: units

       The units defined in the YANG module.

       :rtype: str

    .. property:: default

       The default value as defined in the YANG module.

       :rtype: str, int

    .. property:: mandatory

       Identifies whether the item is required (mandatory) in the YANG module.

       :rtype: bool


    .. Reviewed by PLM 20221005
    .. Reviewed by TechComms 20221005
    """

    __slots__ = ("_model")
    _attributes = (
        "module",
        "namespace",
        "yang_type",
        "units",
        "default",
        "mandatory",
    )

    def __init__(self, model):
        self._model = model

    def __dir__(self):
        retval = super(self.__class__, self).__dir__()
        for attr in self._attributes:
            if hasattr(self, attr):
                retval.append(attr)
        return retval

    def __getattr__(self, attr):
        if attr == "module":
            return self._model.prefix
        if attr == "namespace":
            return self._model.namespace
        if attr == "yang_type" and self._model.yang_type:
            return SchemaType(self._model.yang_type)
        if attr == "units" and self._model.units:
            return self._model.units
        if attr == "default" and self._model.default:
            assert isinstance(self._model.default, (list, str))
            default = self._model.default
            res = []
            if isinstance(default, str):
                default = [default]
            for d in default:
                if self._model.yang_type.json_name() == "boolean":
                    if d == "true":
                        res.append(True)
                    if d == "false":
                        res.append(False)
                    continue
                if self._model.yang_type.json_name() in INTEGRAL_LEAF_TYPE:
                    try:
                        res.append(int(d))
                    except ValueError:
                        pass
                    continue
                res.append(d)
            if isinstance(self._model.default, str):
                res = res[0]
            return res
        if attr == "mandatory":
            if self._model.data_def_stm in (
                self._model.StatementType.leaf_,
                self._model.StatementType.leaf_list_,
                self._model.StatementType.choice_,
                self._model.StatementType.anydata_,
                self._model.StatementType.anyxml_,
            ):
                return True if self._model.mandatory else False
        raise make_exception(
            pysros_err_attr_object_has_no_attribute,
            obj=self.__class__.__name__,
            attribute=attr
        )

    def __eq__(self, other):
        attrs_eq = lambda self, other, attr: getattr(self, attr, None) == getattr(other, attr, None)
        return (
            self.__class__ is other.__class__
            and all(attrs_eq(self, other, attr) for attr in self._attributes)
        )


class SchemaType:
    """Type information for YANG node. Resolves to a YANG base type.

    .. property:: range

       The range defined in YANG.

       :rtype: str

    .. container::

       .. property:: union_members

          Base YANG types that form part of the YANG union.

          :rtype: tuple

    .. Reviewed by PLM 20221005
    .. Reviewed by TechComms 20221005
    """
    __slots__ = ("_yang_type")
    _attributes = (
        "range",
        "union_members",
    )

    def __init__(self, yt):
        self._yang_type = yt

    def __dir__(self):
        retval = super(self.__class__, self).__dir__()
        for attr in self._attributes:
            if hasattr(self, attr):
                retval.append(attr)
        return retval

    def __getattr__(self, attr):
        if attr == "range":
            if self._yang_type.json_name() in DECIMAL_LEAF_TYPE:
                return self._yang_type.yang_range
        if attr == "union_members":
            if isinstance(self._yang_type, YangUnion):
                return tuple(SchemaType(yt) for yt in self._yang_type)
        raise make_exception(
            pysros_err_attr_object_has_no_attribute,
            obj=self.__class__.__name__,
            attribute=attr
        )

    def __str__(self):
        return self._yang_type.json_name()

    def __repr__(self):
        return self._yang_type.json_name()

    def __eq__(self, other):
        attrs_eq = lambda self, other, attr: getattr(self, attr, None) == getattr(other, attr, None)
        return (
            self.__class__ is other.__class__
            and self._yang_type.json_name() == other._yang_type.json_name()
            and all(attrs_eq(self, other, attr) for attr in self._attributes)
        )


class Wrapper:
    """Common functionality to support wrappers that describe the YANG structure from
    the SR OS schema.

    .. warning::
       Instance of this class SHOULD NOT be created by user of pysros library.

    .. Reviewed by PLM 20220923
    .. Reviewed by TechComms 20210712
    """

    __slots__ = ('_data', '_model', '_annotations')

    def __init__(self, value, annotations=None):
        self._data = value
        self._model = None
        self._annotations = annotations

    def __getattr__(self, attr):
        if attr == "_data":
            raise AttributeError()
        return getattr(self._data, attr)

    def __setattr__(self, attr, value):
        if attr == '_data' and not hasattr(self, '_data'):
            self._check_data_type(value)
            object.__setattr__(self, attr, value)
            return
        elif attr == '_model' and getattr(self, '_model', None) is None:
            object.__setattr__(self, attr, value)
            return
        elif attr == '_annotations' or attr == 'annotations':
            object.__setattr__(self, '_annotations', value)
            return
        raise make_exception(
            pysros_err_attr_is_read_only,
            obj=self.__class__.__name__,
            attribute=attr
        )

    def __delattr__(self, attr):
        raise make_exception(
            pysros_err_attr_cannot_be_deleted,
            obj=self.__class__.__name__,
            attribute=attr
        )

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False
        if self._data != other._data:
            return False
        lhs, rhs = getattr(self, 'annotations', None), getattr(other, 'annotations', None)

        if lhs == rhs:
            return True
        is_empty = lambda ann: bool(not ann or (isinstance(ann, list) and not any(ann)))
        lhs_empty, rhs_empty = is_empty(lhs), is_empty(rhs)
        return lhs_empty and rhs_empty

    @property
    def annotations(self):
        if self._annotations is None:
            self._annotations = Annotations()
        return self._annotations

    @property
    def data(self):
        return self._data

    __module__ = 'pysros.wrappers'

    def __dir__(self):
        result = ['__module__', 'annotations', 'schema', 'data', '__getattr__', '__slots__']
        old_dir = object.__dir__(self)
        for item in dir(self.data):
            if not item.startswith('_') or item in old_dir:
                result.append(item)
        return result

    @property
    def schema(self):
        return Schema(self._model) if self._model else None

    @schema.setter
    def schema(self, _):
        raise make_exception(
            pysros_err_attr_is_read_only,
            obj=self.__class__.__name__,
            attribute='schema'
        )

    @staticmethod
    @abstractmethod
    def _check_data_type(value):
        pass

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        s = StringIO()
        s.write(self.__class__.__name__)
        s.write('(')
        s.write(repr(self._data))
        anns = self._annotations
        if anns is not None:
            if (isinstance(anns, Annotations) and anns) or (isinstance(anns, list) and any(anns)):
                s.write(', annotations = ')
                s.write(repr(self._annotations))
        s.write(')')
        return s.getvalue()

    @classmethod
    def _with_model(cls, data, model, *, annotations=None):
        obj = cls(data)
        obj._model = model
        obj._annotations = annotations
        return obj


def forward_methods(C, *methods):
    def make_forward(method):
        op = getattr(operator, method, None)
        if op is not None:
            # solves most methods like __add__, __and__, , but not special cases
            def forwad(self, *args):
                return op(self.data, *args)
            return forwad
        if method.startswith('__r'):
            n_method = "__" + method[3:]
            if hasattr(operator, n_method):  # reverse operator __radd__, __rsub__, __rxor, etc.
                def forward(self, arg):
                    return getattr(operator, n_method)(arg, self.data)
                return forward

        # some methods (eg. __len__) are not defined in operator
        def forward(self, *args):
            return getattr(self.data, method)(*args)
        return forward
    for method in methods:
        setattr(C, method, make_forward(method))


class Container(Wrapper):
    """YANG container data structure node wrapper.

    A YANG container in the pySROS data structure behaves in the same way as a Python dict,
    where the value of a field can be obtained using the field name.

    :ivar data: dictionary containing the fields data

    .. Reviewed by PLM 20210630
    .. Reviewed by TechComms 20210712
    """

    __slots__ = ()

    def __init__(self, value=None, *,  annotations=None):
        super().__init__(value if value is not None else {}, annotations)

    @classmethod
    def _check_data_type(cls, value):
        if not isinstance(value, dict):
            raise make_exception(
                pysros_err_unsupported_type_for_wrapper,
                wrapper_name=cls.__name__
            )


forward_methods(
    Container,
    '__contains__',
    '__delitem__',
    '__getitem__',
    '__iter__',
    '__len__',
    '__reversed__',
    '__setitem__',
    '__or__',
    '__ror__',
)


class Action(Container):
    """YANG Action data structure node wrapper.

    Action in the pySROS data structure behaves as Container, but
    it it returned by convert action.

    :ivar data: dictionary containing the fields data
    """

    pass


class LeafList(Wrapper):
    """
    YANG leaf-list data structure node wrapper.

    A YANG leaf-list in the pySROS data structure behaves in the same way as a Python list, where the separate
    values can be obtained using an index.

    :ivar data: list containing the values

    .. Reviewed by PLM 20210630
    .. Reviewed by TechComms 20210712
    """

    __slots__ = ()

    def __init__(self, value=None, *, annotations=None):
        super().__init__(value if value is not None else [], annotations)

    @classmethod
    def _check_data_type(cls, values):
        if not isinstance(values, list):
            raise make_exception(
                pysros_err_unsupported_type_for_wrapper,
                wrapper_name=cls.__name__
            )
        for value in values:
            Leaf._check_data_type(value)

    @property
    def annotations(self):
        """
        List of YANG annotations (metadata) applied to the LeafList.

        Applying YANG annotations to a :py:class:`LeafList` uses the same constructs
        detailed in the :py:class:`Annotations` section.  When applied to LeafLists the
        :py:attr:`.annotations` property is a Python list.  Each entry in the LeafList
        has an entry in the :py:attr:`.annotations` list.

        Each entry in the LeafList may have zero or more annotations.  This list of
        annotations per entry is represented as an instance of the
        :py:class:`Annotations` class.  Each instance of :py:class:`Annotations`
        has zero or more instances of the :py:class:`Annotation`.

        This concept is better illustrated through examples.

        .. note:: The example below will not validate against the Nokia YANG models but
                  is provided to detail the implementation and usage of YANG annotations
                  attached to :py:class:`LeafList` classes.


        Consider the following LeafList:

        .. code-block:: python
           :name: annotations-leaflist-initiallist

           >>> my_leaflist
           LeafList(['one', 'two', 'three'])

        The list of annotations attached to this :py:class:`LeafList` is stored in the ``annotations``
        attribute:

        .. code-block:: python
           :name: annotations-leaflist-initiallist-annotations

           >>> my_leaflist.annotations
           []

        This is an empty list as there are no annotations defined against any entry in the :py:class:`LeafList`.
        Annotations will now be added to the first entry and the last entry of the list.  The first entry will
        have two annotations and the last one:

        .. code-block:: python
           :caption: Example showing annotations in :py:class:`LeafList` objects
           :name: annotations-leaflist-set-annotations

           >>> my_leaflist.annotations = [
           ...     Annotations([
           ...         Annotation(key='key1', data='value1'),
           ...         Annotation(key='key2', data='value2')
           ...     ]),
           ...     Annotations([]),
           ...     Annotations([
           ...         Annotation(key='key3', data='value3')
           ...     ])
           ... ]
           >>> my_leaflist.annotations
           [
               Annotations([
                   Annotation(key='key1', data='value1'),
                   Annotation(key='key2', data='value2')]),
               Annotations([]),
               Annotations([
                   Annotation(key='key3', data='value3')
               ])
           ]

        .. Reviewed by PLM 20231003

        """
        if self._annotations is None:
            self._annotations = [Annotations() for _ in self._data]
        return self._annotations


forward_methods(
    LeafList,
    '__add__', '__contains__', '__delitem__', '__getitem__', '__iadd__',
    '__imul__', '__iter__', '__len__', '__mul__', '__reversed__', '__rmul__',
    '__setitem__'
)


class Leaf(Wrapper):
    """YANG leaf data structure node wrapper.

    Depending on the base (root) YANG type of the node, this object wraps a str, int, or bool.

    :ivar data: python object corresponding to the value

    .. Reviewed by PLM 20210630
    .. Reviewed by TechComms 20210713
    """

    __slots__ = ()

    @staticmethod
    def _check_data_type(data):
        if not isinstance(data, (str, int, bool, _Empty)) and data is not None:
            raise make_exception(
                pysros_err_unsupported_type_for_wrapper,
                wrapper_name='Leaf'
            )

    def __bool__(self):
        return bool(self.data)


forward_methods(Leaf,
                '__lt__',
                '__le__',
                '__ge__',
                '__gt__',
                '__abs__',
                '__add__',
                '__and__',
                '__ceil__',
                '__contains__',
                '__divmod__',
                '__float__',
                '__floor__',
                '__floordiv__',
                '__getitem__',
                '__getnewargs__',
                '__index__',
                '__int__',
                '__invert__',
                '__iter__',
                '__len__',
                '__lshift__',
                '__mod__',
                '__mul__',
                '__neg__',
                '__or__',
                '__pos__',
                '__pow__',
                '__radd__',
                '__rand__',
                '__rdivmod__',
                '__rfloordiv__',
                '__rlshift__',
                '__rmod__',
                '__rmul__',
                '__ror__',
                '__round__',
                '__rpow__',
                '__rrshift__',
                '__rshift__',
                '__rsub__',
                '__rtruediv__',
                '__rxor__',
                '__sub__',
                '__truediv__',
                '__trunc__',
                '__xor__',)
del forward_methods

_sentinel = object()

class _AnnotationAttr:
    """Internal class for handling attributes of Annotation."""
    def __set_name__(self, owner, name):
        self.public_name = name
        self.private_name = '_' + name

    def __set__(self, obj, value):
        if self.private_name != '_data':
            setattr(obj, '_model', None)
        if self.private_name == '_key' and not isinstance(value, str):
            raise make_exception(pysros_err_annotation_type_key)
        setattr(obj, self.private_name, value)

    def __get__(self, obj, objtype=None):
        value = getattr(obj, self.private_name)
        if value is None:
            raise make_exception(
                pysros_err_attr_object_has_no_attribute,
                obj=obj.__class__.__name__,
                attribute=self.public_name)
        return value

    def __delete__(self, obj):
        if getattr(obj, self.private_name) is None or self.private_name == '_key':
            raise make_exception(
                pysros_err_attr_object_has_no_attribute,
                obj=obj.__class__.__name__,
                attribute=self.public_name
            )
        setattr(obj, self.private_name, None)

@total_ordering
class Annotation:
    """
    Wrapper for an individual annotation (YANG metadata) associated with an elements
    in the data structure.  An element in the data structure may have more than one
    :py:class:`Annotation`.  The :py:class:`Annotations` class is used to define this.

    Allows operator to provide an arbitrary key/value pair to any element in the pySROS
    data structure.  The data in the :py:class:`Annotation` must exist in the nodes YANG
    schema for the :py:meth:`pysros.management.Datastore.set` and
    :py:meth:`pysros.management.Connection.convert` methods to function.

    :ivar key: The name of the YANG annotation statement.  This may be in simple form
               or in namespace qualified form, for example: ``comment`` or ``nokia-attr:comment``.

    :ivar data: The value of the annotation.

    :ivar module: The YANG module name in which the YANG annotation exists.

    :ivar namespace: The YANG module namespace in which the YANG annotation exists.

    .. Reviewed by PLM 20231003

    """

    __slots__ = ('_key', '_data', '_module', '_namespace', '_model')
    key = _AnnotationAttr()
    """The name of the YANG annotation statement.  This may be in simple form
    or in namespace qualified form, for example: `comment` or `nokia-attr:comment`.
    
    .. Reviewed by PLM 20231003
    """
    data = _AnnotationAttr()
    """The value of the annotation.
    
    .. Reviewed by PLM 20231003
    """
    module = _AnnotationAttr()
    """The YANG module name in which the YANG annotation exists.
    
    .. Reviewed by PLM 20231003
    """
    namespace = _AnnotationAttr()
    """The YANG module namespace in which the YANG annotation exists.
    
    .. Reviewed by PLM 20231003
    """

    def __init__(self, key, data=None, *, module=None, namespace=None):
        if not isinstance(key, str):
            raise make_exception(pysros_err_annotation_type_key)

        if ':' in key:
            nmodule, key = key.split(':', 1)
            if module is not None and module != nmodule:
                raise make_exception(pysros_err_annotation_invalid_module)
            module = nmodule

        self._key = key
        self._data = data
        self._module = module
        self._namespace = namespace
        self._model = None

    @classmethod
    def _with_model(cls, *arg, model, **kwarg):
        res = cls(*arg, **kwarg)
        res._model = model
        return res

    @property
    def schema(self):
        """YANG schema supporting information associated with elements in the data structure (Read only).

        .. Reviewed by PLM 20231003

        """
        return self._model and Schema(self._model)

    def __repr__(self):
        comma = False
        stream = StringIO()
        stream.write(self.__class__.__name__)
        stream.write('(')
        for attr in self.__slots__:
            if attr.startswith('_'):
                attr = attr[1:]
            a = getattr(self, attr, _sentinel)
            if a != _sentinel:
                if comma:
                    stream.write(', ')
                comma = True
                stream.write(f'{attr}={getattr(self, attr)!r}')
        stream.write(')')
        return stream.getvalue()

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return NotImplemented
        return (
            self._key == other._key
            and self._data == other._data
            and self._module == other._module
            and self._namespace == other._namespace
        )

    def __lt__(self, other):
        def is_less(a,b):
            return False if (b is None) else (a is None or a < b)

        if self.__class__ is not other.__class__:
            return NotImplemented

        if self._key != other._key: return is_less(self._key, other._key)
        if self._module != other._module: return is_less(self._module, other._module)
        if self._namespace != other._namespace: return is_less(self._namespace, other._namespace)
        return is_less(self._data, other._data)


class Annotations(collections.UserList):
    """Annotations class provides wrapper for instances of the Annotation class that must be supplied
    in the same way as Python *list*.

    The annotations (there may be more than one) on a particular YANG node are defined in the ``annotations``
    parameter attached to a :py:class:`Leaf`, :py:class:`Container` or :py:class:`LeafList`.
    The ``annotations`` parameter is a class of type :py:class:`.Annotations` (with
    an 's').  This is a list of :py:class:`.Annotation` (without an 's') class instances.

    An :py:class:`.Annotation` in pySROS is treated in a similar way as any
    other YANG structure such as a :py:class:`.Leaf`.  A :py:class:`.Annotation`
    class wrapper encodes the structures required to define and use a YANG-modeled annotation.  Unlike other
    wrappers, because a YANG-modeled annotation can be in a different YANG namespace from the node it is
    attached to, additional information is needed.

    The :py:meth:`pysros.management.Connection.convert` method and the :py:meth:`pysros.management.Datastore.set`
    method support the :py:class:`.Annotation` class.  This means that, as with other data in
    pySROS, annotations can be entered without providing the details of the YANG module and these methods
    derive the correct YANG model information using the YANG schema learned from the specific router.

    .. note::

       Applying annotations to :py:class:`LeafList` objects requires special consideration.  Please refer to
       the :py:attr:`.LeafList.annotations` section for more information.

    The following example obtains a :py:class:`.Leaf` from the device, in this instance the system name, and
    adds a configuration comment to it which it then sets back to the router using the
    :py:meth:`pysros.management.Datastore.set` method.  The example assumes that a
    :py:class:`pysros.management.Connection` object called ``connection_object`` has already been created:

    .. code-block:: python
       :caption: Set a configuration comment on the system name obtained from the device
       :name: annotations-set-comment-on-sysname

       >>> from pysros.wrappers import Annotations, Annotation
       >>> path = '/nokia-conf:configure/system/name'
       >>> system_name = connection_object.running.get(path)
       >>> system_name
       Leaf('sros')
       >>> system_name.annotations = [Annotation('comment', 'This is my comment')]
       >>> connection_object.candidate.set(path, system_name)

    The following example obtains a :py:class:`.Leaf` from the device and adds a configuration comment to it in
    the native Python data structure.  It then uses the :py:meth:`pysros.management.Connection.convert` method
    to query the known YANG schema for the given router and add the YANG model information to the
    :py:class:`.Annotation`.  The example assumes that a :py:class:`pysros.management.Connection` object called
    ``connection_object`` has already been created:

    .. code-block:: python
       :caption: Use convert to identify and complete YANG specific attributes in the object
       :name: annotations-convert-comment-on-sysname
       :emphasize-lines: 6-10,15,20,25-26

       >>> from pysros.wrappers import Annotations, Annotation
       >>> path = '/nokia-conf:configure/system/name'
       >>> system_name = connection_object.running.get(path)
       >>> system_name
       Leaf('sros')
       >>> system_name.annotations
       Annotations([])
       >>> system_name.annotations.append(Annotation('comment', 'This is my configuration comment'))
       >>> system_name.annotations
       Annotations([Annotation(key='comment', data='This is my configuration comment')])
       >>> connection_object.convert(path,
       ...                           system_name,
       ...                           source_format='pysros',
       ...                           destination_format='xml')
       '<nokia-conf:name xmlns:nokia-conf="urn:nokia.com:sros:ns:yang:sr:conf" xmlns:nokia-attr="urn:nokia.com:sros:ns:yang:sr:attributes" nokia-attr:comment="This is my configuration comment">sros</nokia-conf:name>'
       >>> connection_object.convert(path,
       ...                           system_name,
       ...                           source_format='pysros',
       ...                           destination_format='json')
       '{"@nokia-conf:name": {"nokia-attributes:comment": "This is my configuration comment"}, "nokia-conf:name": "sros"}'
       >>> augmented_system_name = connection_object.convert(path,
       ...                                                   system_name,
       ...                                                   source_format='pysros',
       ...                                                   destination_format='pysros')
       >>> augmented_system_name.annotations
       Annotations([Annotation(key='comment', data='This is my configuration comment', module='nokia-attributes', namespace='urn:nokia.com:sros:ns:yang:sr:attributes')])

    This next example takes a valid NETCONF RPC encoded in XML that inserts a new item into a user-ordered YANG
    list.  The well-known ``operation`` annotation.  Using the :py:class:`.Annotations` and
    :py:class:`.Annotation` classes, this RPC can be converted into native the native Python (pySROS)
    data structure, manipulated if required, and sent to the router.  The example assumes that a
    :py:class:`pysros.management.Connection` object called ``connection_object`` has already been created.

    The initial configuration of the user-ordered list is as follows:

    .. code-block:: text
       :caption: Initial state of the user-ordered list
       :name: annotations-user-ordered-list-before

       (gl)[/configure policy-options policy-statement "example"]
       A:admin@sros# info
           entry-type named
           named-entry "one" {
               action {
                   action-type accept
               }
           }
           named-entry "three" {
               action {
                   action-type accept
               }
           }
           default-action {
               action-type reject
           }


    .. code-block:: python
       :caption: Convert NETCONF operations using system defined annotations
       :name: annotations-system-defined
       :emphasize-lines: 8,19

        >>> from pysros.wrappers import Annotation, Annotations
        >>> from pysros.management import connect
        >>> my_xml = \"\"\"
        ... <configure xmlns="urn:nokia.com:sros:ns:yang:sr:conf" xmlns:yang="urn:ietf:params:xml:ns:yang:1" xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0">
        ...     <policy-options>
        ...         <policy-statement>
        ...             <name>example</name>
        ...             <named-entry nc:operation="merge" yang:insert="after" yang:key="[entry-name='one']">
        ...                 <entry-name>two</entry-name>
        ...                 <action>
        ...                     <action-type>accept</action-type>
        ...                 </action>
        ...             </named-entry>
        ...         </policy-statement>
        ...     </policy-options>
        ... </configure>
        ... \"\"\"
        >>> my_rpc = connection_object.convert('/', my_xml, source_format='xml', destination_format='pysros')
        {'configure': Container({'policy-options': Container({'policy-statement': {'example': Container({'name': Leaf('example'), 'named-entry': OrderedDict([('two', Container({'entry-name': Leaf('two'), 'action': Container({'action-type': Leaf('accept')})}, annotations = Annotations([Annotation(key='operation', data='merge', module='ietf-netconf', namespace='urn:ietf:params:xml:ns:netconf:base:1.0'), Annotation(key='insert', data='after', module='yang', namespace='urn:ietf:params:xml:ns:yang:1'), Annotation(key='key', data="[entry-name='one']", module='yang', namespace='urn:ietf:params:xml:ns:yang:1')])))])})}})})}
        >>> connection_object.candidate.set('/nokia-conf:configure', my_rpc['configure'])


    .. code-block:: text
       :caption: Resulting state of the user-ordered list
       :name: annotations-user-ordered-list-after
       :emphasize-lines: 9-13

       (gl)[/configure policy-options policy-statement "example"]
       A:admin@sros# info
           entry-type named
           named-entry "one" {
               action {
                   action-type accept
               }
           }
           named-entry "two" {
               action {
                   action-type accept
               }
           }
           named-entry "three" {
               action {
                   action-type accept
               }
           }
           default-action {
               action-type reject
           }

    The :py:class:`.Annotations` class behaves in a similar way to a native Python list and many of the
    functions that operate on a Python list will also function on the :py:class:`.Annotations` class.
    Examples include :py:func:`index` (which returns the numerical location in the list of the requested
    item), :py:func:`remove` (which deletes a specific item from a list) and :py:func:`append` (which
    adds an item to the end of a list).

    .. note:: When using :py:func:`index` and :py:func:`remove` the exact matching :py:class:`.Annotation`
              must be provided including all associated parameters.

    A number of additional helper methods are provided for searching through :py:class:`.Annotations` and
    selecting specific :py:class:`.Annotation` entries.  These are described below:

    .. Reviewed by PLM 20231001

    """

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return NotImplemented
        return self.data == other.data

    def index_annotation(self, key, data=None, *, module=None, namespace=None):
        """Return the index to searched :py:attr:`.Annotation.key`.  An exception
        is raised if the :py:class:`.Annotation` is not found.

        :param key: Name of the :py:class:`Annotation`.  This should match the
                    :py:attr:`.Annotation.key` in order to yield a successful match.
                    If more than one :py:class:`Annotation` exists with the same
                    :py:attr:`.Annotation.key` other parameters must be provided.
        :type key: str

        :param data: The value (:py:attr:`.Annotation.data`) of the
                     :py:class:`Annotation`.
        :type data: str, optional

        :param module: The YANG module name that :py:class:`Annotation` exists in.
        :type module: str, optional

        :param namespace: The YANG module namespace that the :py:class:`Annotation`
                          exists in.
        :type namespace: str, optional

        :raises ValueError: A matching :py:class:`Annotation` is not found.

        .. note:: The example below will not validate against the Nokia YANG models but
                  is provided to detail the implementation and usage of the :py:class:`Annotations`
                  class.

        .. code-block:: python
           :caption: :py:meth:`.Annotations.index_annotation` example
           :name: annotations-index-annotation
           :emphasize-lines: 3-4

           >>> myleaf.annotations
           Annotations([Annotation(key='key1', data='data1'), Annotation(key='key2', data='data2'), Annotation(key='key3', data='data3')])
           >>> myleaf.annotations.index_annotation('key2')
           1

        .. code-block:: python
           :caption: :py:meth:`.Annotations.index_annotation` example handling a failed match
           :name: annotations-index-annotation-failed-match
           :emphasize-lines: 3-8

           >>> myleaf.annotations
           Annotations([Annotation(key='key1', data='data1'), Annotation(key='key2', data='data2'), Annotation(key='key3', data='data3')])
           >>> try:
           ...   myleaf.annotations.index_annotation('foo')
           ... except ValueError as error:
           ...   print("Failed to find Annotation:", error)
           ...
           Failed to find Annotation: Annotation not found

        .. Reviewed by PLM 20231003

        """

        def match(required, given):
            return (required is None) or (required == given)

        if not isinstance(key, str):
            raise make_exception(pysros_err_annotation_type_key)

        if ':' in key and module == None:
            module, key = key.split(':', 1)

        idx = None
        for i, d in enumerate(self.data):
            if (not match(key, d._key)
                    or not match(data, d._data)
                    or not match(module, d._module)
                    or not match(namespace, d._namespace)):
                continue

            if idx is not  None:
                raise make_exception(pysros_err_annotation_too_many_instances)

            idx = i

        if idx is None:
            raise  make_exception(pysros_err_annotation_not_found)
        return idx

    def get(self, annotation):
        """Return the :py:class:`.Annotation` (without an ‘s’) from the :py:class:`.Annotations`
        (with an ‘s’) class (list) that matches the provided :py:class:`.Annotation`.

        The match must be identical, that is, all parameters inside the :py:class:`.Annotation`
        class *must match exactly*.

        :param annotation: Annotation object to match in the :py:class:`.Annotations` list.
        :type annotation: :py:class:`.Annotation`

        :raises ValueError: Error if the requested :py:class:`.Annotation` is not found.

        .. note:: The example below does not validate against the Nokia YANG models but
          is provided to detail the implementation and usage of the :py:class:`Annotations`
          class.

        .. code-block:: python
           :caption: :py:meth:`.Annotations.get` example
           :name: annotations-get
           :emphasize-lines: 3-4

           >>> myleaf.annotations
           Annotations([Annotation(key='key1', data='data1'), Annotation(key='key2', data='data2'), Annotation(key='key3', data='data3')])
           >>> myleaf.annotations.get(Annotation(key='key3', data='data3'))
           Annotation(key='key3', data='data3')

        .. code-block:: python
           :caption: :py:meth:`.Annotations.get` example handling a failed match
           :name: annotations-get-failed-match
           :emphasize-lines: 3-8

           >>> myleaf.annotations
           Annotations([Annotation(key='key1', data='data1'), Annotation(key='key2', data='data2'), Annotation(key='key3', data='data3')])
           >>> try:
           ...   myleaf.annotations.get('key1')
           ... except ValueError as error:
           ...   print("Failed to find Annotation:", error)
           ...
           Failed to find Annotation: 'key1' is not in list

        .. Reviewed by PLM 20231003
        .. Reviewed by TechComms 20231009

        """
        idx = self.data.index(annotation)
        return self.data[idx]

    def get_annotation(self, key, data=None, *, module=None, namespace=None):
        """
        Return the :py:class:`.Annotation` (without an ‘s’) from the :py:class:`.Annotations`
        (with an ‘s’) class (list) that matches the provided :py:attr:`.Annotation.key`.
        An exception is raised if the :py:class:`Annotation` is not found.

        :param key: Name of the :py:class:`Annotation`.  This should match the
                    :py:attr:`.Annotation.key` in order to yield a successful match.
                    If more than one :py:class:`Annotation` exists with the same
                    :py:attr:`.Annotation.key` other parameters must be provided.
        :type key: str

        :param data: The value (:py:attr:`.Annotation.data`) of the
                     :py:class:`Annotation`.
        :type data: str, optional

        :param module: The YANG module name that :py:class:`Annotation` exists in.
        :type module: str, optional

        :param namespace: The YANG module namespace that the :py:class:`Annotation`
                          exists in.
        :type namespace: str, optional

        :raises ValueError: A matching :py:class:`Annotation` is not found.

        .. note:: The example below does not validate against the Nokia YANG models but
                  is provided to detail the implementation and usage of the :py:class:`Annotations`
                  class.

        .. code-block:: python
           :caption: :py:meth:`.Annotations.get_annotation` example
           :name: annotations-get-annotation
           :emphasize-lines: 3-4

           >>> myleaf.annotations
           Annotations([Annotation(key='key1', data='data1'), Annotation(key='key2', data='data2'), Annotation(key='key3', data='data3')])
           >>> myleaf.annotations.get_annotation('key2')
           Annotation(key='key2', data='data2')

        .. Reviewed by PLM 20231003
        .. Reviewed by TechComms 20231009

        """
        idx = self.index_annotation(key, data, module=module, namespace=namespace)
        return self.data[idx]


    def remove_annotation(self, key, data=None, *, module=None, namespace=None):
        """Remove the :py:class:`.Annotation` object for the searched
        :py:attr:`.Annotation.key` from the :py:class:`.Annotations` class (list).
        An exception is raised if the :py:class:`.Annotation` is not found.

        :param key: Name of the :py:class:`Annotation`.  This should match the
                    :py:attr:`.Annotation.key` in order to yield a successful match.
                    If more than one :py:class:`Annotation` exists with the same
                    :py:attr:`.Annotation.key` other parameters must be provided.
        :type key: str

        :param data: The value (:py:attr:`.Annotation.data`) of the
                     :py:class:`Annotation`.
        :type data: str, optional

        :param module: The YANG module name that :py:class:`Annotation` exists in.
        :type module: str, optional

        :param namespace: The YANG module namespace that the :py:class:`Annotation`
                          exists in.
        :type namespace: str, optional

        :raises ValueError: A matching :py:class:`Annotation` is not found.

        .. note:: The example below does not validate against the Nokia YANG models but
                  is provided to detail the implementation and usage of the :py:class:`Annotations`
                  class.

        .. code-block:: python
           :caption: :py:meth:`.Annotations.remove_annotation` example
           :name: annotations-remove-annotation
           :emphasize-lines: 3-4

           >>> myleaf.annotations
           Annotations([Annotation(key='key1', data='data1'), Annotation(key='key2', data='data2'), Annotation(key='key3', data='data3')])
           >>> myleaf.annotations.remove_annotation('key2')
           >>> myleaf.annotations
           Annotations([Annotation(key='key1', data='data1'), Annotation(key='key3', data='data3')])

        .. Reviewed by PLM 20231003
        .. Reviewed by TechComms 20231009

        """
        idx = self.index_annotation(key, data, module=module, namespace=namespace)
        del self.data[idx]

    def __repr__(self):
        return f"Annotations({self.data!r})"




# Copyright 2021-2024 Nokia

import copy
from collections import defaultdict, OrderedDict
from itertools import chain
from typing import DefaultDict, Dict, Optional, Set, Union, List
from decimal import Decimal

from .errors import *
from .errors import InvalidPathError, make_exception
from .identifier import Identifier, LazyBindModule, Wildcard
from .model import AModel, BuildingModel, Model, StorageConstructionModel, YangVersion
from .model_path import ModelPath
from .path_utils import get_path
from .model_walker import DataModelWalker, ModelWalker
from .request_data import COMMON_NAMESPACES
from .tokenizer import yang_parser
from .yang_type import (DECIMAL_LEAF_TYPE, INTEGRAL_TYPES_MAX,
                        INTEGRAL_TYPES_MIN, Bits, Enumeration,
                        IdentityRef, LeafRef, PrimitiveType,
                        UnresolvedIdentifier, YangType, YangTypeBase,
                        YangUnion, should_be_buildin, type_from_name)


class YangHandler:
    """Handler for yang processing."""
    TAGS_WITH_MODEL = (
        "container", "list", "leaf", "typedef", "module", "submodule", "uses",
        "leaf-list", "import", "identity", "notification", "rpc", "belongs-to",
        "input", "output", "choice", "case", "deviate", "action",
    )
    TAGS_FORCE_MODULE = (
        "grouping", "identity", "typedef", "uses", "augment",
        "deviation", "type", "base", "refine",
    )
    TAGS_ARG_IS_IDENTIFIER = ("base", "type", )
    TAGS_ARG_IS_YANG_PATH_NOT_ABSOLUTE_SCHEMA_ID = ("path", )
    assert not ({
        "config", "default", "fraction-digits", "length", "mandatory",
        "max-elements", "min-elements", "must", "range", "status", "type",
        "unique", "units",
    } & set(TAGS_WITH_MODEL)), "Substmt of deviate can not have model"
    TAGS_SHOULD_BE_PROCESSED = (
        "action",
        "anydata",
        "anyxml",
        "argument",
        "augment",
        "base",
        "belongs-to",
        "bit",
        "case",
        "choice",
        "config",
        "contact",
        "container",
        "default",
        "description",
        "deviate",
        "deviation",
        "enum",
        # "error-app-tag",
        # error-message",
        # extension",
        "feature",
        "fraction-digits",
        "grouping",
        "identity",
        "if-feature",
        "import",
        "include",
        "input",
        "key",
        "leaf",
        "leaf-list",
        "length",
        "list",
        "mandatory",
        "max-elements",
        "min-elements",
        "modifier",
        "module",
        # "must",
        "namespace",
        "notification",
        "ordered-by",
        "organization",
        "output",
        "path",
        "pattern",
        "position",
        "prefix",
        "presence",
        "range",
        "reference",
        "refine",
        "require-instance",
        "revision",
        "revision-date",
        "rpc",
        "status",
        "submodule",
        "type",
        "typedef",
        "unique",
        "units",
        "uses",
        "value",
        # "when",
        "yang-version",
        "yin-element",
    )

    def __init__(self, builder: "ModelBuilder", root: BuildingModel):
        self.builder = builder
        self.root = root
        self.path = [root]
        self.full_path = []
        self.grouping_depth = 0
        self.in_type = 0
        self.derived_identities: Dict[Identifier, Set[Identifier]] = {}
        self.include_stack = [{}]
        self.yang_version_stack = [YangVersion()]
        self.module = None
        self._prefix = [None]
        self.ignore_depth = 0

    def enter(self, name, arg):
        if self.ignore_depth or (name not in self.TAGS_SHOULD_BE_PROCESSED and ":" not in name):
            self.ignore_depth += 1
            return
        self.full_path.append(name)

        if name in ("input", "output", ):
            arg = name
        if name == "case" and not arg:
            arg = "unnamed"
        if name == 'typedef':
            new_id = self.get_identifier(arg, name in self.TAGS_FORCE_MODULE)
            assert not new_id.is_lazy_bound()
            new = BuildingModel(new_id, BuildingModel.StatementType.typedef_, None, self.yang_version)
            self.path.append(new)
        elif name in self.TAGS_WITH_MODEL:
            new_id = self.get_identifier(arg, name in self.TAGS_FORCE_MODULE)
            new = BuildingModel(new_id, BuildingModel.StatementType[name.replace("-", "_") + "_"], self.model, self.yang_version)
            self.path.append(new)
        elif name in ("augment", "deviation", "refine", ):
            new_id = self.get_identifier(name, name in self.TAGS_FORCE_MODULE)
            if self.model.data_def_stm == Model.StatementType.uses_:
                new = BuildingModel(new_id, BuildingModel.StatementType[name + "_"], self.model, self.yang_version)
                new.target_path = get_path(arg, absolute_path_allowed=False, relative_path_allowed=True, key_allowed=False, prefix_resolver=self.resolve_prefix)
            else:
                # global augment or deviation
                new = BuildingModel(new_id, BuildingModel.StatementType[name + "_"], None, self.yang_version)
                (self.builder.augments if name == "augment" else self.builder.deviations).append(new)
                new.target_path = get_path(arg, absolute_path_allowed=True, relative_path_allowed=False, key_allowed=False, prefix_resolver=self.resolve_prefix)
            self.path.append(new)
        elif name == "grouping":
            new_id = self.get_identifier(arg, name in self.TAGS_FORCE_MODULE)
            new = BuildingModel(new_id, BuildingModel.StatementType[name.replace("-", "_") + "_"], None, self.yang_version)
            self.path.append(new)
            assert new_id not in self.builder.groupings
            self.builder.groupings[new_id] = new
        elif ":" in name:
            new_id = Identifier.from_model_string(name)
            new = BuildingModel(new_id, BuildingModel.StatementType.extended, self.model, self.yang_version)
            new.arg = f"{self.module}:{arg}"
            new.nsmap = self.prefixModuleMapping
            self.path.append(new)
        else:
            if name in self.TAGS_ARG_IS_IDENTIFIER:
                if name == "type" and should_be_buildin(arg):
                    arg = Identifier.builtin(arg)
                else:
                    arg = self.get_identifier(arg, name in self.TAGS_FORCE_MODULE, force_allow_yang_string=True)
            elif name in self.TAGS_ARG_IS_YANG_PATH_NOT_ABSOLUTE_SCHEMA_ID:
                arg = get_path(arg, absolute_path_allowed=True, relative_path_allowed=True, key_allowed=True, key_expect_path=True, prefix_resolver=self.resolve_prefix, default_module=self.module)
            self.model.blueprint.append((True, (name, arg)))

        handle_name = f'handle_early_{name.replace("-", "_DASH_")}'
        if hasattr(self, handle_name):
            handler = getattr(self, handle_name)
            handler(arg)

    def leave(self, name):
        if self.ignore_depth:
            self.ignore_depth -= 1
            return
        if name == "grouping":
            self.grouping_depth -= 1
        elif name == "typedef":
            self.construct(self.model)
            self.builder.types_to_resolve[self.model.name] = TypeDefModel(self.model)
        elif name in ("action", "rpc"):
            for child in self.model.children:
                if child.data_def_stm == BuildingModel.StatementType.input_:
                    break
            else:
                self.enter("input", None)
                self.leave("input")
            for child in self.model.children:
                if child.data_def_stm == BuildingModel.StatementType.output_:
                    break
            else:
                self.enter("output", None)
                self.leave("output")

        popped = self.full_path.pop()
        assert popped == name
        if name in (*self.TAGS_WITH_MODEL, "grouping", "augment", "deviation", "refine",) or ":" in name:
            self.path.pop()
        else:
            self.model.blueprint.append((False, name))

    def construct(self, model=None):
        model = model or self.root
        self.path.append(model)
        self.full_path.append(model.data_def_stm.name.replace("_", "-")[:-1])
        for instruction in model.blueprint:
            if instruction[0]:
                self.full_path.append(instruction[1][0])
                if self.full_path[-1] == "type":
                    self.in_type += 1
                self.construct_enter(instruction[1])
            else:
                if self.full_path.pop() == "type":
                    self.in_type -= 1
        for child in self.model.children:
            self.construct(child)
        self.path.pop()
        self.full_path.pop()

    def construct_enter(self, instruction):
        name, arg = instruction
        handle_name = f'handle_{name.replace("-", "_DASH_")}'
        if hasattr(self, handle_name):
            handler = getattr(self, handle_name)
            handler(arg)

    def get_identifier(self, s, force_module=False, force_allow_yang_string=False):
        assert not force_module or self.module is not None
        if self.module is not None:
            module = LazyBindModule() if (self.grouping_depth and not force_module) else self.module
            return Identifier.from_yang_string(s, module, self.prefixModuleMapping)
        parts = s.split(":")
        if force_allow_yang_string and len(parts) > 1:
            return Identifier(parts[0], parts[1])
        return Identifier.builtin(s)

    @property
    def model(self):
        return self.path[-1]

    @property
    def model_type(self):
        return self.model.data_def_stm

    @property
    def prefixModuleMapping(self):
        return self.include_stack[-1]

    @property
    def yang_version(self)->YangVersion:
        return self.yang_version_stack[-1]

    @yang_version.setter
    def yang_version(self, name):
        self.yang_version_stack[-1].value = name

    @property
    def last_yang_type(self):
        return self.model.yang_type[-1] if isinstance(self.model.yang_type, YangUnion) else self.model.yang_type

    def handle_identity(self, identifier):
        assert self.model.data_def_stm == BuildingModel.StatementType.identity_

    def handle_base(self, ident):
        if self.in_type:
            assert isinstance(self.last_yang_type, IdentityRef)
            self.last_yang_type.add_base(ident)
        elif self.model.data_def_stm == BuildingModel.StatementType.identity_:
            self.model.identity_bases.append(ident)

    def handle_path(self, arg: str):
        if self.in_type:
            self.last_yang_type.set_path(arg)

    def handle_type(self, identifier: str):
        if not isinstance(self.model.yang_type, YangUnion):
            assert self.model.yang_type is None
            self.model.yang_type = type_from_name(identifier)
        elif (identifier != Identifier.builtin('union')):
            self.model.yang_type.append(type_from_name(identifier))

    def handle_units(self, arg: str):
        assert self.model.units is None
        self.model.units = arg

    def handle_namespace(self, arg: str):
        assert self.model.namespace is None
        self.model.namespace = arg

    def handle_default(self, arg: str):
        if self.model.data_def_stm == Model.StatementType.leaf_list_:
            if self.model.default is None:
                self.model.default = [arg]
            else:
                self.model.default.append(arg)
        else:
            assert self.model.default is None
            self.model.default = arg

    def handle_mandatory(self, arg: str):
        assert self.model.mandatory is None
        self.model.mandatory = arg

    def handle_range(self, arg: str):
        assert isinstance(self.last_yang_type, (PrimitiveType, UnresolvedIdentifier))
        assert self.last_yang_type.yang_range is None
        self.last_yang_type.yang_range = arg

    def handle_fraction_DASH_digits(self, arg: str):
        assert isinstance(self.last_yang_type, (PrimitiveType, UnresolvedIdentifier))
        assert self.last_yang_type.fraction_digits is None
        self.last_yang_type.fraction_digits = int(arg)

    def handle_length(self, arg: str):
        assert isinstance(self.last_yang_type, (PrimitiveType, UnresolvedIdentifier))
        assert self.last_yang_type.length is None
        self.last_yang_type.length = arg

    def handle_status(self, arg: str):
        self.model.status = arg

    def handle_enum(self, value: str):
        if self.in_type:
            assert isinstance(self.last_yang_type, (Enumeration, UnresolvedIdentifier, )), f"expected Enumeration, got {self.last_yang_type}"
            self.last_yang_type.add_enum(value)

    def handle_bit(self, value: str):
        if self.in_type:
            assert isinstance(self.last_yang_type, (Bits, UnresolvedIdentifier, ))
            self.last_yang_type.add_bit(value)

    def handle_value(self, value: str):
        if self.in_type:
            assert isinstance(self.last_yang_type, (Enumeration, UnresolvedIdentifier, ))
            self.last_yang_type.set_last_enum_value(int(value))

    def handle_presence(self, _arg):
        self.model.presence_container = True

    def handle_ordered_DASH_by(self, arg):
        if arg == "user":
            self.model.user_ordered = True

    def handle_early_import(self, arg):
        self.builder.register_yang(arg)

    def handle_early_include(self, arg):
        self.include_stack.append({})
        self.yang_version_stack.append(YangVersion())
        self.builder.perform_parse(arg, self)
        self.yang_version_stack.pop()
        self.include_stack.pop()

    def handle_key(self, arg):
        self.model.local_keys.extend(arg.split())

    def handle_early_module(self, arg):
        self.module = arg

    def handle_early_grouping(self, _arg):
        self.grouping_depth += 1

    def handle_early_prefix(self, arg):
        if self.full_path[-2] in ("belongs-to", ):
            assert arg not in self.prefixModuleMapping
            self.prefixModuleMapping[arg] = self.module
        if self.full_path[-2] in ("import", "module", ):
            assert arg not in self.prefixModuleMapping
            self.prefixModuleMapping[arg] = self.model.name.name

    def handle_early_yang_DASH_version(self, arg):
        self.yang_version = arg

    def handle_config(self, arg: str):
        if arg not in ('false', 'true'):
            raise make_exception(pysros_err_invalid_config)
        self.model.config = (arg == 'true')

    def resolve_prefix(self, prefix: str):
        return self.prefixModuleMapping[prefix]


def _dummy_getter(filename):
    assert False, "Missing getter"


class ModelBuilder:
    """API for walking model tree."""
    def __init__(self, yang_getter=_dummy_getter, ns_map={}, sros=True):
        self.root = BuildingModel(
            "root",
            BuildingModel.StatementType["container_"],
            None,
            YangVersion.ver1_1
        )
        self.metadata = []
        self.types_to_resolve = dict()
        self.groupings = dict()
        self.resolved_types = dict()
        self.all_yangs = set()
        self.parsed_yangs = set()
        self.files_to_parse = []
        self.augments = []
        self.deviations = []
        self.registered_modules = {}
        self._ns_map = ns_map
        self.yang_getter = yang_getter
        self._sros = sros

    def get_module_content(self, yang_name):
        if yang_name in self.registered_modules:
            return self.registered_modules[yang_name]
        return self.yang_getter(yang_name)

    def register_yangs(self, yang_names):
        for name in yang_names:
            self.register_yang(name)

    def register_yang(self, yang_name):
        assert isinstance(yang_name, str)
        if yang_name not in self.all_yangs:
            self.all_yangs.add(yang_name)
            self.files_to_parse.append(yang_name)

    def DEBUG_parse_File(self, module_name, f):
        if module_name not in self.all_yangs:
            self.all_yangs.add(module_name)
            self.parsed_yangs.add(module_name)
        if module_name not in self._ns_map:
            self._ns_map[module_name] = f"NS:{module_name}"
        yang_parser(f, YangHandler(self, self.root))

    def DEBUG_register_module(self, module_name, f):
        assert module_name not in self.registered_modules
        self.registered_modules[module_name] = f

    def process_all_yangs(self):
        while self.files_to_parse:
            self.perform_parse(self.files_to_parse.pop())
        self.resolve()

    def resolve(self):
        self.resolve_groupings()
        self.resolve_augments()
        self.resolve_deviations()
        self.build_blueprints()
        self.resolve_typedefs()
        self.resolve_type_ranges()
        self.resolve_identities()
        self.resolve_leafrefs()
        self.resolve_config()
        self.resolve_namespaces()
        self.delete_blueprints()
        self.resolve_metadata()
        self.resolve_metadata_exceptions()
        self.convert_model()

    def walk_models(self, cb):
        self.root.recursive_walk(cb)
        for a in self.metadata:
            a.recursive_walk(cb)

    def perform_parse(self, yang_name, yang_handler=None):
        if yang_name in self.parsed_yangs:
            return
        self.parsed_yangs.add(yang_name)

        yang_parser(
            self.get_module_content(yang_name),
            yang_handler or YangHandler(self, self.root)
        )

    def set_correct_types(self, m: BuildingModel):
        if isinstance(m.yang_type, (UnresolvedIdentifier, YangUnion)):
            tdm = resolve_typedefs_deep(TypeDefModel(m), self.resolved_types)
            m.yang_type = tdm.yang_type
            m.default = tdm.default
            m.units = tdm.units
        #RFC6020: If the type referenced by the leaf-list has a default value, it has no effect in the leaf-list.
        if m.data_def_stm == AModel.StatementType.leaf_list_ and m.default and m.yang_version == YangVersion.ver1_0:
            m.default = None

    def resolve_typedefs(self):
        for key, value in self.types_to_resolve.items():
            assert key not in self.resolved_types
            self.resolved_types[key] = resolve_typedefs_deep(value, self.types_to_resolve)
        self.walk_models(self.set_correct_types)

    def set_correct_range(self, yt: YangTypeBase):
        assert yt.json_name() != "decimal64" or yt.fraction_digits
        if isinstance(yt, YangUnion):
            for t in yt:
                self.set_correct_range(t)

        if not yt or not yt.json_name() in (*DECIMAL_LEAF_TYPE, "string", "binary"):
            return
        if yt.json_name() in ("string", "binary") and not yt.length:
            return

        full_range = False if (yt.yang_range or yt.json_name() in ("string", "binary")) else True
        work_range = yt.length if yt.json_name() in ("string", "binary") else yt.yang_range

        if full_range:
            work_range = "min..max"
        if yt.json_name() == "decimal64":
            min_val = str(-2**63 + 1)
            max_val = str(2**63 - 1)
            min_val = min_val[:-yt.fraction_digits] + '.' + min_val[-yt.fraction_digits:]
            max_val = max_val[:-yt.fraction_digits] + '.' + max_val[-yt.fraction_digits:]
        elif yt.json_name() in ("string", "binary"):
            min_val = str(0)
            max_val = str(2**64 - 1)
        else:
            min_val = INTEGRAL_TYPES_MIN[yt.json_name()]
            max_val = INTEGRAL_TYPES_MAX[yt.json_name()]

        work_range = work_range.replace("min", min_val)
        work_range = work_range.replace("max", max_val)

        if full_range:
            yt.yang_range = work_range
            return

        # merge adjacent subranges, such as "0|1..100" to "0..100" or "0..10|11..20" to "0..20"
        def process_subrange(subrange: str):
            nonlocal yt
            map_f = float if yt.json_name() == "decimal64" else int
            i_subrange = list(map(map_f, subrange.split("..")))
            assert len(i_subrange) in (1, 2)
            return i_subrange if len(i_subrange) == 2 else 2 * i_subrange

        subranges = work_range.split("|")
        res_subranges = [process_subrange(subranges[0])]
        for subrange in subranges[1:]:
            i_subrange = process_subrange(subrange)
            if res_subranges[-1][1] == i_subrange[0] - 1:
                res_subranges[-1][1] = i_subrange[1]
            else:
                res_subranges.append(i_subrange)

        res_subranges = [subr if subr[0] != subr[1] else [subr[0]] for subr in res_subranges]
        work_range = "|".join(
            "..".join(map(str, subr)) for subr in res_subranges
        )
        if yt.json_name() in ("string", "binary"):
            yt.length = work_range
        else:
            yt.yang_range = work_range

    def resolve_type_ranges(self):
        def setter(m: BuildingModel):
            if isinstance(m.yang_type, YangTypeBase):
                self.set_correct_range(m.yang_type)
        self.walk_models(setter)

    def resolve_identities(self):
        # first step creates dict, where we find for each identity
        # its all identities, that are directly derived from it.
        directly_derived: DefaultDict[Identifier, Set[Identifier]] = defaultdict(set)

        def find_derived(m: BuildingModel):
            if m.data_def_stm != BuildingModel.StatementType.identity_:
                return
            for b in m.identity_bases:
                if m.status != 'obsolete':
                    directly_derived[b].add(m.name)

        self.walk_models(find_derived)
        # second step adds all direct and indirect derived identities together

        def add_derived_recursive(result_set, id):
            got = directly_derived.get(id, set())
            toUpdate = got - result_set
            result_set.update(got)
            for c in toUpdate:
                add_derived_recursive(result_set, c)

        derived: Dict[Identifier, Set(Identifier)] = {}
        for id in directly_derived.keys():
            derived[id] = set()
            add_derived_recursive(derived[id], id)

        # last step iterates through all identityrefs and set possible values
        def identityref_set_value(m: BuildingModel):
            if m.yang_type is None:
                return
            if type(m.yang_type) is IdentityRef:
                m.yang_type.set_values(derived, self._ns_map)
            elif type(m.yang_type) is YangUnion:
                for st in m.yang_type:
                    if type(st) is IdentityRef:
                        st.set_values(derived, self._ns_map)

        self.walk_models(identityref_set_value)

    def resolve_leafrefs(self):
        def replace_leafrefs(m: BuildingModel):
            if not isinstance(m.yang_type, LeafRef):
                return

            w = DataModelWalker(m, self._sros)
            assert m.data_def_stm in (
                BuildingModel.StatementType.leaf_,
                BuildingModel.StatementType.leaf_list_
            )

            while isinstance(w.current.yang_type, LeafRef):
                w.current.yang_type.path.move_walker(w, default_module=w.current.prefix)
            m.yang_type = w.current.yang_type
        self.walk_models(replace_leafrefs)

    def resolve_groupings(self):
        module = None
        def resolve_grouping(m: BuildingModel):
            if m.data_def_stm == BuildingModel.StatementType.uses_:
                grouping: BuildingModel = self.groupings[m.name]
                grouping.recursive_walk(resolve_grouping)
                uses_children, m.children = m.children, []
                for child in grouping.children:
                    child.deepcopy(m)
                self.process_uses_substmts(m, uses_children)
                m.annihilate()
                return False
        def resolve_common(m: BuildingModel, module: str):
            assert module is not LazyBindModule()
            def resolve_unresolved_prefix(m: BuildingModel):
                if m.prefix is LazyBindModule():
                    m.prefix = module
            grouping: BuildingModel = self.groupings[m.name]
            uses_children, m.children = m.children, []
            for child in grouping.children:
                child.deepcopy(m).recursive_walk(resolve_unresolved_prefix)
            self.process_uses_substmts(m, uses_children)
            m.annihilate()
            return False
        def resolve_data(m: BuildingModel):
            nonlocal module
            if m.data_def_stm == BuildingModel.StatementType.module_:
                module = m.name.name
            if m.data_def_stm == BuildingModel.StatementType.uses_:
                return resolve_common(m, module)
        def resolve_augment(m: BuildingModel):
            if m.data_def_stm == BuildingModel.StatementType.uses_:
                module = augment.prefix
                return resolve_common(m, module)
        for grouping in self.groupings.values():
            grouping.recursive_walk(resolve_grouping)
        self.root.recursive_walk(resolve_data)
        for augment in self.augments:
            augment.recursive_walk(resolve_augment)

    def process_uses_substmts(self, uses: BuildingModel, substmts: List[BuildingModel]):
        for m in substmts:
            if m.data_def_stm == BuildingModel.StatementType.refine_:
                w = ModelWalker(uses, self._sros)
                m.target_path.move_walker(w, Wildcard())
                self.resolve_deviations_replace(w.current, m)
            if m.data_def_stm == BuildingModel.StatementType.augment_:
                w = ModelWalker(uses, self._sros)
                m.target_path.move_walker(w, Wildcard())
                for i in m.children:
                    i.deepcopy(w.current)

    def process_augment(self, m: BuildingModel):
        if m.data_def_stm == BuildingModel.StatementType.augment_:
            w = ModelWalker.path_parse(self.root, self._sros, m.target_path, True)
            for i in m.children:
                i.deepcopy(w.current)

    def resolve_augments(self):
        current = self.augments
        while current:
            remaining = []
            for augment in current:
                w = ModelWalker(self.root, self._sros)
                try:
                    w.go_to(augment.target_path)
                    node = w.current
                except InvalidPathError:
                    remaining.append(augment)
                    continue
                for i in augment.children:
                    i.deepcopy(node)
            if len(remaining) >= len(current):
                raise make_exception(pysros_err_unresolved_augment)
            current = remaining

    def filter_blueprint(self, function, blueprint):
        filtering_depth = 0
        for instruction in blueprint:
            if filtering_depth:
                filtering_depth += 1 if instruction[0] else -1
                continue
            if instruction[0]:
                if function(instruction[1]):
                    yield instruction
                else:
                    filtering_depth = 1
            else:
                yield instruction

    def resolve_deviations(self):
        for deviation in self.deviations:
            w = ModelWalker(self.root, self._sros)
            w.go_to(deviation.target_path)
            for deviate in deviation.children:
                if deviate.name.name == "add":
                    w.current.blueprint.extend(deviate.blueprint)
                if deviate.name.name == "delete":
                    for instruction in deviate.blueprint:
                        if instruction[0]:
                            w.current.blueprint = list(self.filter_blueprint(lambda b: b[0:1] != instruction[1][0:1], w.current.blueprint))
                if deviate.name.name == "replace":
                    self.resolve_deviations_replace(w.current, deviate)
                if deviate.name.name == "not-supported":
                    w.current.delete_from_parent(quiet=False)

    def resolve_deviations_replace(self, target: BuildingModel, deviate: BuildingModel):
        depth = 0
        for instruction in deviate.blueprint:
            if not depth:
                target.blueprint = list(self.filter_blueprint(lambda b: b[0] != instruction[1][0], target.blueprint))
            depth += 1 if instruction[0] else -1
        target.blueprint = list(target.blueprint) + deviate.blueprint

    def resolve_namespaces(self):
        def ns_set(m: BuildingModel):
            if m.namespace or m.name.prefix not in self._ns_map.keys():
                return
            m.namespace = self._ns_map[m.name.prefix]
        for root_child in chain(self.root.children, self.metadata):
            root_child.recursive_walk(ns_set)

    def build_blueprints(self):
        handler = YangHandler(self, self.root)
        handler.construct()
        for a in self.metadata:
            handler.construct(a)

    def delete_blueprints(self):
        def deleter(m: BuildingModel):
            del m.blueprint
        self.walk_models(deleter)

    def resolve_config(self):
        def false_setter(m: BuildingModel):
            m.config = False

        def resolver(m: BuildingModel):
            if not m.config:
                m.recursive_walk(false_setter)
                return False

        self.walk_models(resolver)

    def resolve_metadata(self):
        def resolver(m: BuildingModel):
            if m.data_def_stm == Model.StatementType.extended:
                m.name = Identifier.from_yang_string(m.name.model_string, None, m.nsmap)

                if m.name == Identifier("ietf-yang-metadata", "annotation"):
                    m.delete_from_parent()
                    m.data_def_stm = Model.StatementType.annotate_
                    m.name = Identifier.from_model_string(m.arg)
                    m.namespace = self._ns_map[m.name.prefix]
                    self.metadata.append(m)
                    return False
        self.root.recursive_walk(resolver)

    def resolve_metadata_exceptions(self):
        new = BuildingModel(Identifier("ietf-netconf", "operation"), BuildingModel.StatementType.annotate_, None, YangVersion.ver1_0)
        new.namespace = COMMON_NAMESPACES["ncbase"]
        type = Enumeration()
        type.add_enums("merge", "replace", "create", "delete", "remove")
        new.yang_type = type
        self.metadata.append(new)

    def convert_model(self):
        new_root = StorageConstructionModel("root", BuildingModel.StatementType["container_"], None)
        w = DataModelWalker(self.root, self._sros)

        stack = [new_root]

        def enter_fnc(w: DataModelWalker):
            nonlocal stack
            new = StorageConstructionModel(w.get_name(), w.get_dds(), stack[-1])
            if isinstance(w.current.yang_type, YangUnion):
                w.current.yang_type.deduplicate()
            new.yang_type = w.current.yang_type
            new.units = w.current.units
            new.namespace = w.current.namespace
            new.default = w.current.default
            new.mandatory = w.current.mandatory
            new.status = w.current.status
            new.presence_container = w.current.presence_container
            new.user_ordered = w.current.user_ordered
            new.local_keys = w.current.local_keys
            new.target_path = w.current.target_path
            new.identity_bases = w.current.identity_bases
            new.config = w.current.config
            stack.append(new)

        def leave_fnc(w: DataModelWalker):
            nonlocal stack
            stack.pop()

        w.recursive_walk(enter_fnc=enter_fnc, leave_fnc=leave_fnc)
        self.root = Model(new_root._storage, new_root._index, None)


class TypeDefModel:
    __slots__ = ("yang_type", "default", "units")

    def __init__(self, arg: Union[YangType, Model, BuildingModel], default: Optional[str] = None, units: Optional[str] = None) -> None:
        assert isinstance(arg, (YangTypeBase, AModel))
        if isinstance(arg, AModel):
            self.yang_type = arg.yang_type
            self.default = arg.default
            self.units = arg.units
        else:
            self.yang_type = arg
            self.default = default
            self.units = units

    def __eq__(self, other):
        if not isinstance(other, self.__class__) or self.__slots__ != other.__slots__:
            return False
        return all(getattr(self, name) == getattr(other, name) for name in self.__slots__)


def merge_typedef_ranges(parent_range: str, child_range: str):
    """
    Vertically merge typedef integer ranges across typedef parent and child
    - keep child's integer range values (non-'min' or 'max')
    - replace child's 'min'/'max' strings with integer min or max values if present in parent
    - return child's ranges as-is if parent has no range defined or child has no 'max' or 'min'
    examples:
        - parent typedef range "1..200", child type range "min..100", result "1..100"
        - parent typedef range "1..200", child type range "100..max", result "100..200"
        - parent typedef range is None, child type range "100..max", result "100..max"
        - parent typedef range is '1..200', child type range "50..150", result "50..150"
    """
    if not parent_range:
        return child_range

    min_val = parent_range.split("|")[0].split("..")[0]
    child_range = child_range.replace("min", min_val)
    max_val = parent_range.split("|")[-1].split("..")[-1]
    child_range = child_range.replace("max", max_val)

    return child_range


def resolve_typedefs_shallow(t: TypeDefModel, typedefs: Dict[Identifier, TypeDefModel]):
    assert isinstance(t.yang_type, YangTypeBase)
    yang_type, default, units = t.yang_type, t.default, t.units
    while (type(yang_type) is UnresolvedIdentifier and
            yang_type.identifier in typedefs):
        r_model = typedefs[yang_type.identifier]
        u_range = yang_type.yang_range
        frac_d = yang_type.fraction_digits
        length = yang_type.length
        enums = yang_type.enums
        if any((u_range, frac_d, length, enums, )):
            yang_type = copy.copy(r_model.yang_type)
            if u_range:
                yang_type.yang_range = merge_typedef_ranges(yang_type.yang_range, u_range)
            if frac_d:
                yang_type.fraction_digits = frac_d
            if length:
                yang_type.length = length
            if enums:
                if isinstance(yang_type, (Enumeration, Bits, )):
                    yang_type.enums = OrderedDict(filter(lambda e: e[0] in enums, yang_type.enums.items()))
                else:
                    yang_type.enums = enums
        else:
            yang_type = r_model.yang_type
        default = default or r_model.default
        units = units or r_model.units

    assert not isinstance(yang_type, UnresolvedIdentifier), f"Cannot resolve type {yang_type}"
    assert isinstance(yang_type, YangTypeBase)
    return TypeDefModel(
        copy.copy(yang_type),
        default,
        units,
    )


def resolve_typedefs_deep(m: TypeDefModel, typedefs: Dict[Identifier, TypeDefModel]):
    m = resolve_typedefs_shallow(m, typedefs)
    if not isinstance(m.yang_type, YangUnion):
        return m

    u_yang_type, u_default, u_units = YangUnion(), None, None
    for sub in m.yang_type:
        tdm = resolve_typedefs_deep(TypeDefModel(sub), typedefs)
        if type(tdm.yang_type) == YangUnion:
            for t in tdm.yang_type:
                u_yang_type.append(t)
        else:
            u_yang_type.append(tdm.yang_type)
        u_default = tdm.default
        u_units = tdm.units
    m.yang_type = u_yang_type
    m.default = m.default or u_default
    m.units = m.units or u_units
    return m



from .model_path import ModelPath
from .identifier import Identifier
from .singleton import _Singleton
from .model_walker import ModelWalker

import re
import functools
import contextlib
from typing import Optional, List, Iterable, Tuple, Dict, Union

class LevelUp(metaclass=_Singleton):
    def __str__(self):
        return f"{self.__class__.__name__}()"

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return self is other


class _Path:
    def __init__(self):
        self._path: List[Identifier] = []
        self._keys: List[Dict[Identifier, Union[str, _Path]]] = []
        self.default_module = None

    def append(self, i: Identifier):
        self._path.append(i)
        self._keys.append({})

    def append_level_up(self):
        self.append(LevelUp())

    def move_walker(self, w: ModelWalker, default_module=None):
        for name, keys in zip(self._path, self._keys):
            if name is LevelUp():
                w.go_to_parent()
            elif name.is_lazy_bound():
                w.go_to_child(Identifier(default_module or w.current.prefix, name.name))
            else:
                w.go_to_child(name)
            if not w.is_root:
                w.local_keys.update(keys)

    @property
    def top_keys(self):
        return self._keys[-1]


class AbsolutePath(_Path):
    def move_walker(self, w: ModelWalker, *args, **kwargs):
        while not w.is_root:
            w.go_to_parent()
        super().move_walker(w, *args, **kwargs)


class RelativePath(_Path):
    pass

"""
leafref:
===================================================================
   leafref-specification =
                         ;; these stmts can appear in any order
                         path-stmt
                         [require-instance-stmt]

   path-stmt           = path-keyword sep path-arg-str stmtend

   path-arg-str        = < a string that matches the rule >
                         < path-arg >
===================================================================

augment:
===================================================================
   augment-arg-str     = < a string that matches the rule >
                         < augment-arg >

   augment-arg         = absolute-schema-nodeid
===================================================================

deviation:
===================================================================
   deviation-arg-str   = < a string that matches the rule >
                         < deviation-arg >

   deviation-arg       = absolute-schema-nodeid
===================================================================

uses:
===================================================================
   refine-arg-str      = < a string that matches the rule >
                         < refine-arg >

   refine-arg          = descendant-schema-nodeid

   uses-augment-arg-str = < a string that matches the rule >
                          < uses-augment-arg >

   uses-augment-arg    = descendant-schema-nodeid
===================================================================

json-instance-path:
===================================================================
   An "instance-identifier" value is encoded as a string that is
   analogical to the lexical representation in XML encoding; see
   Section 9.13.2 in [RFC7950].  However, the encoding of namespaces in
   instance-identifier values follows the rules stated in Section 4,
   namely:

   o  The leftmost (top-level) data node name is always in the
      namespace-qualified form.

   o  Any subsequent data node name is in the namespace-qualified form
      if the node is defined in a module other than its parent node, and
      the simple form is used otherwise.  This rule also holds for node
      names appearing in predicates.
===================================================================
"""

"""
path-arg:
  leafref

absolute-schema-nodeid:
  augment
  deviation

descendant-schema-nodeid:
  uses-refine
  uses-augment

json-instance-path:
  pysros
"""

"""
   path-arg            = absolute-path / relative-path

   absolute-path       = 1*("/" (node-identifier *path-predicate))

   relative-path       = 1*("../") descendant-path

   descendant-path     = node-identifier
                         [*path-predicate absolute-path]

   path-predicate      = "[" *WSP path-equality-expr *WSP "]"

   path-equality-expr  = node-identifier *WSP "=" *WSP path-key-expr

   path-key-expr       = current-function-invocation *WSP "/" *WSP
                         rel-path-keyexpr

   rel-path-keyexpr    = 1*(".." *WSP "/" *WSP)
                         *(node-identifier *WSP "/" *WSP)
                         node-identifier

   node-identifier     = [prefix ":"] identifier

   current-function-invocation = current-keyword *WSP "(" *WSP ")"

   current-keyword          = %s"current"

absolute-path:
    /a
    /a/b
    /prefix:a
    /a/b[key=current()/../c]
    /a/b[key=current()/../c][key2=current()/../c][key3=current()/../c]
    /a/b[prefix:key=current()/../c]
    /a/b[  prefix:key  =  current  (  )  /  ..  /  c]
    #not in grammar, but probably still correct
    /  a  /  b

relative-path:
    ../a
    ../a/b
    ../../../a/b
    ../prefix:a
    ../a/b[key=current()/../c]
    ../a/b[key=current()/../c][key2=current()/../c][key3=current()/../c]
    ../a/b[prefix:key=current()/../c]
    ..  /  a  /  b  [  prefix:key  =  current  (  )  /  ..  /  c]

"""

"""
   ;; Schema Node Identifiers

   schema-nodeid       = absolute-schema-nodeid /
                         descendant-schema-nodeid

   absolute-schema-nodeid = 1*("/" node-identifier)

   descendant-schema-nodeid =
                         node-identifier
                         [absolute-schema-nodeid]

   node-identifier     = [prefix ":"] identifier

absolute-schema-nodeid:
    /a
    /a/b/c
    /prefix:a

descendant-schema-nodeid:
    a
    a/b/c
    prefix:a
"""

"""
   ;; Instance Identifiers

   instance-identifier = 1*("/" (node-identifier
                                 [1*key-predicate /
                                  leaf-list-predicate /
                                  pos]))

   key-predicate       = "[" *WSP key-predicate-expr *WSP "]"

   key-predicate-expr  = node-identifier *WSP "=" *WSP quoted-string

   leaf-list-predicate = "[" *WSP leaf-list-predicate-expr *WSP "]"

   leaf-list-predicate-expr = "." *WSP "=" *WSP quoted-string

   pos                 = "[" *WSP positive-integer-value *WSP "]"

   quoted-string       = (DQUOTE string DQUOTE) / (SQUOTE string SQUOTE)

   node-identifier     = [prefix ":"] identifier

Instance Identifiers:
    /ex:system/ex:services/ex:ssh
    /ex:system/ex:services/ex:ssh/ex:port
    /ex:system/ex:user[ex:name='fred']
    /ex:system/ex:user[ex:name='fred']/ex:type
    /ex:system/ex:server[ex:ip='192.0.2.1'][ex:port='80']
    /ex:system/ex:service[ex:name='foo'][ex:enabled='']
    /ex:system/ex:services/ex:ssh/ex:cipher[.='blowfish-cbc']
    /ex:stats/ex:port[3]
    /   ex:stats   /   ex:port   [   3   ]

Json instance identifier:
    /mod:system/services/ssh
    /mod:system/user[name='fred']
    /mod:system/services/mod2:ssh

Pysros instance identifier:
    /system/services/ssh
    /system/user[name='fred']
    /system/services/ssh
"""

SLASH_RE = r'\s*(?P<SLASH>[/])\s*'
BRACKET_OPEN_RE = r'\s*(?P<BRACKET_OPEN>\[)\s*'
BRACKET_CLOSE_RE = r'\s*(?P<BRACKET_CLOSE>\])\s*'
EQ_RE = r'\s*(?P<EQ>=)\s*'
_NAME_RE = r"""[a-zA-Z_][a-zA-Z0-9_.-]*"""
NAME_RE = fr"""(?P<NAME>{_NAME_RE})"""
IDENTIFIER_RE = fr"""(?P<IDENTIFIER>{_NAME_RE}:{_NAME_RE})"""
CURRENT_NODE_RE = r"(?P<CURRENT_NODE>[.])"
PARENT_NODE_RE = r"(?P<PARENT_NODE>[.][.])"
QUOTED_STR_RE = r"""(?P<QUOTED_STR>(?:["](?:[^\\"]|[\\][\\nt"])*["])|(?:['](?:[^'])*[']))"""
CURRENT_FNC_RE = r"\s*(?P<CURRENT_FNC>current\s*[(]\s*[)])\s*"


class Tokenizer:
    def __init__(self, s: str):
        self.s = s
        self.pos = 0
        self.current_match = None

    def consume(self, regex: Iterable[str]):
        if self.finished:
            return False
        r = re.compile(fr"(?:{'|'.join(regex)})")
        self.current_match = r.match(self.s, self.pos)
        if self.current_match is None:
            return False
        self.pos = self.current_match.end()
        return True

    def match(self, regex: Iterable[str]):
        if self.finished:
            return False
        r = re.compile(fr"(?:{'|'.join(regex)})")
        match = r.match(self.s, self.pos)
        if match is None:
            return False
        return True

    @property
    def kind(self):
        return self.current_match.lastgroup

    @property
    def token(self):
        return self.current_match.group(self.current_match.lastindex)

    @property
    def finished(self):
        return self.pos >= len(self.s)


_ = (SLASH_RE, BRACKET_OPEN_RE, BRACKET_CLOSE_RE, EQ_RE, CURRENT_FNC_RE, PARENT_NODE_RE, CURRENT_NODE_RE, IDENTIFIER_RE, NAME_RE, QUOTED_STR_RE, )

"""
    absolute_path
        path_segment+

    relative_path
        (PARENT_NODE / SLASH) * / ( node_segment / absolute_path ? ) ?

    node
        (NAME|IDENTIFIER)
    path_segment
        SLASH / node_segment
    node_segment
        node / [ key_segment* ]
        node /* for schema node id */
    key_segment
        BRACKET_OPEN / node / EQ / path_key_val / BRACKET_CLOSE
    path_key_val
        CURRENT_FNC / SLASH /relative_path
        QUOTED_STR /* for json instance path */

"""


class Parser:
    def __init__(self, s: str, *, key_allowed, absolute_path_allowed, relative_path_allowed, prefix_resolver=None, key_expect_path=None):
        assert key_expect_path is not None or not key_allowed
        self.tokenizer = Tokenizer(s)
        self.key_allowed = key_allowed
        self.absolute_path_allowed = absolute_path_allowed
        self.relative_path_allowed = relative_path_allowed
        self.key_expect_path = key_expect_path
        if prefix_resolver is not None:
            self.prefix_identifier_resolver = prefix_resolver
        else:
            self.prefix_identifier_resolver = lambda x: x

    def parse(self):
        if self.peek_slash():
            self.path = AbsolutePath()
            assert self.absolute_path_allowed
        else:
            self.path = RelativePath()
            assert self.relative_path_allowed
            while self.try_levelup():
                if not self.tokenizer.finished:
                    self.slash()
                    assert not self.tokenizer.finished
                self.path.append_level_up()

            if not self.tokenizer.finished:
                component = self.identifier_with_keys()
                self.path.append(component[0])
                for name, val in component[1]:
                    self.path.top_keys[name] = val

        while not self.tokenizer.finished:
            component = self.path_component()
            self.path.append(component[0])
            for name, val in component[1]:
                self.path.top_keys[name] = val

    def try_levelup(self):
        return self.tokenizer.consume((PARENT_NODE_RE, ))

    def peek_slash(self):
        return self.tokenizer.match((SLASH_RE, ))

    def try_slash(self):
        return self.tokenizer.consume((SLASH_RE, ))

    def peek_bracket_open(self):
        return self.tokenizer.match((BRACKET_OPEN_RE, ))

    def slash(self):
        assert self.tokenizer.consume((SLASH_RE, ))
        assert self.tokenizer.kind == "SLASH"

    def path_component(self) -> Tuple[Identifier, List[Tuple[Identifier, str]]]:
        self.slash()
        return self.identifier_with_keys()

    def key_segment(self) -> Tuple[Identifier, str]:
        assert self.tokenizer.consume((BRACKET_OPEN_RE, ))
        name = self.identifier()
        assert self.tokenizer.consume((EQ_RE, ))
        if self.key_expect_path:
            val = self.leaflist_pathj_key()
        else:
            val = self.quoted_str()
        assert self.tokenizer.consume((BRACKET_CLOSE_RE, ))
        return (name, val, )

    def leaflist_pathj_key(self):
        res = RelativePath()
        self.current_fnc()

        while self.try_slash():
            if self.try_levelup():
                res.append_level_up()
            else:
                res.append(self.identifier())

    def identifier_with_keys(self) -> Tuple[Identifier, List[Tuple[Identifier, str]]]:
        identifier = self.identifier()
        keys = []
        if self.key_allowed:
            while self.peek_bracket_open():
                keys.append(self.key_segment())
        return (identifier, keys, )

    def identifier(self) -> Identifier:
        assert self.tokenizer.consume((IDENTIFIER_RE, NAME_RE, ))
        name = self.tokenizer.token
        if self.tokenizer.kind == "IDENTIFIER":
            prefix, name = name.split(":")
            return Identifier(self.prefix_identifier_resolver(prefix), name)
        return Identifier.lazy_bound(name)

    def quoted_str(self):
        assert self.tokenizer.consume((QUOTED_STR_RE, ))
        return self.tokenizer.token[1:-1]

    def current_fnc(self):
        assert self.tokenizer.consume((CURRENT_FNC_RE, ))
        assert self.tokenizer.kind == "CURRENT_FNC"


def parse(w: ModelWalker, *args, **kwargs):
    get_path(*args, **kwargs).move_walker(w)


@functools.wraps(Parser)
def get_path(*args, default_module=None, **kwargs):
    parser = Parser(*args, **kwargs)
    parser.parse()
    res = parser.path
    if default_module:
        res.default_module = default_module
    return res
